@@ -6,11 +6,10 @@
 (*         *       GNU Lesser General Public License Version 2.1        *)
 (************************************************************************)
 
-(* $Id: lib.ml 9488 2007-01-17 11:11:58Z herbelin $ *)
+(* $Id: lib.ml 11784 2009-01-14 11:36:32Z herbelin $ *)
 
 open Pp
 open Util
-open Names
 open Libnames
 open Nameops
 open Libobject
@@ -22,16 +21,18 @@ type node =
   | Leaf of obj
   | CompilingLibrary of object_prefix
   | OpenedModule of bool option * object_prefix * Summary.frozen
+  | ClosedModule  of library_segment
   | OpenedModtype of object_prefix * Summary.frozen
+  | ClosedModtype of library_segment
   | OpenedSection of object_prefix * Summary.frozen
-  | ClosedSection
+  | ClosedSection of library_segment
   | FrozenState of Summary.frozen
 
 and library_entry = object_name * node
 
 and library_segment = library_entry list
 
-type lib_objects =  (identifier * obj) list
+type lib_objects =  (Names.identifier * obj) list
 
 let iter_objects f i prefix =
   List.iter (fun (id,obj) -> f i (make_oname prefix id, obj))
@@ -47,11 +48,18 @@ let subst_objects prefix subst seg =
   in
     list_smartmap subst_one seg
 
+let load_and_subst_objects i prefix subst seg =
+  List.rev (List.fold_left (fun seg (id,obj as node) ->
+    let obj' =  subst_object (make_oname prefix id, subst, obj) in
+    let node = if obj == obj' then node else (id, obj') in
+    load_object i (make_oname prefix id, obj');
+    node :: seg) [] seg)
+
 let classify_segment seg =
   let rec clean ((substl,keepl,anticipl) as acc) = function
     | (_,CompilingLibrary _) :: _ | [] -> acc
     | ((sp,kn as oname),Leaf o) :: stk -> 
-	let id = id_of_label (label kn) in
+	let id = Names.id_of_label (Names.label kn) in
 	  (match classify_object (oname,o) with 
 	     | Dispose -> clean acc stk
 	     | Keep o' -> 
@@ -60,7 +68,11 @@ let classify_segment seg =
 		 clean ((id,o')::substl, keepl, anticipl) stk
 	     | Anticipate o' ->
 		 clean (substl, keepl, o'::anticipl) stk)
-    | (oname,ClosedSection) :: stk -> clean acc stk
+    | (_,ClosedSection _) :: stk -> clean acc stk
+    (* LEM; TODO: Understand what this does and see if what I do is the
+                  correct thing for ClosedMod(ule|type) *)
+    | (_,ClosedModule _) :: stk -> clean acc stk
+    | (_,ClosedModtype _) :: stk -> clean acc stk
     | (_,OpenedSection _) :: _ -> error "there are still opened sections"
     | (_,OpenedModule _) :: _ -> error "there are still opened modules"
     | (_,OpenedModtype _) :: _ -> error "there are still opened module types"
@@ -79,7 +91,7 @@ let segment_of_objects prefix =
    sections, but on the contrary there are many constructions of section
    paths based on the library path. *) 
 
-let initial_prefix = default_library,(initial_path,empty_dirpath)
+let initial_prefix = default_library,(Names.initial_path,Names.empty_dirpath)
 
 let lib_stk = ref ([] : library_segment)
 
@@ -92,35 +104,42 @@ let library_dp () =
    module path and relative section path *)
 let path_prefix = ref initial_prefix
 
+let sections_depth () =
+  List.length (Names.repr_dirpath (snd (snd !path_prefix)))
+
+let sections_are_opened () =
+  match Names.repr_dirpath (snd (snd !path_prefix)) with
+      [] -> false
+    | _ -> true
 
 let cwd () = fst !path_prefix
 
+let current_dirpath sec =
+  Libnames.drop_dirpath_prefix (library_dp ()) 
+    (if sec then cwd () 
+      else Libnames.extract_dirpath_prefix (sections_depth ()) (cwd ()))
+    
 let make_path id = Libnames.make_path (cwd ()) id
 
+let path_of_include () =
+  let dir = Names.repr_dirpath (cwd ()) in
+  let new_dir = List.tl dir in
+  let id = List.hd dir in
+    Libnames.make_path (Names.make_dirpath new_dir) id
 
 let current_prefix () = snd !path_prefix
 
 let make_kn id = 
   let mp,dir = current_prefix () in
-    Names.make_kn mp dir (label_of_id id)
+    Names.make_kn mp dir (Names.label_of_id id)
 
 let make_con id = 
   let mp,dir = current_prefix () in
-    Names.make_con mp dir (label_of_id id)
+    Names.make_con mp dir (Names.label_of_id id)
 
 
 let make_oname id = make_path id, make_kn id
 
-
-let sections_depth () =
-  List.length (repr_dirpath (snd (snd !path_prefix)))
-
-let sections_are_opened () =
-  match repr_dirpath (snd (snd !path_prefix)) with
-      [] -> false
-    | _ -> true
-
-
 let recalc_path_prefix () =
   let rec recalc = function
     | (sp, OpenedSection (dir,_)) :: _ -> dir
@@ -150,18 +169,32 @@ let find_split_p p =
   in
   find !lib_stk
 
-let split_lib sp = 
+let split_lib_gen test = 
   let rec collect after equal = function
-    | ((sp',_) as hd)::before ->
-      	if sp = sp' then collect after (hd::equal) before else after,equal,hd::before
-    | [] -> after,equal,[]
+    | hd::strict_before as before ->
+      	if test hd then collect after (hd::equal) strict_before else after,equal,before
+    | [] as before -> after,equal,before
   in
   let rec findeq after = function
-    | ((sp',_) as hd)::before ->
-      	if sp = sp' then collect after [hd] before else findeq (hd::after) before
-    | [] -> error "no such entry"
-  in 
-  findeq [] !lib_stk
+    | hd :: before ->
+      	if test hd
+	then Some (collect after [hd] before)
+	else (match hd with
+		| (sp,ClosedModule  seg)
+		| (sp,ClosedModtype seg)
+		| (sp,ClosedSection seg) ->
+		    (match findeq after seg with
+		       | None -> findeq (hd::after) before
+		       | Some (sub_after,sub_equal,sub_before) ->
+			   Some (sub_after, sub_equal, (List.append sub_before before)))
+		| _ -> findeq (hd::after) before)
+    | [] -> None
+  in
+    match findeq [] !lib_stk with
+      | None -> error "no such entry"
+      | Some r -> r
+
+let split_lib sp = split_lib_gen (fun x -> (fst x) = sp)
 
 (* Adding operations. *)
 
@@ -170,7 +203,7 @@ let add_entry sp node =
 
 let anonymous_id = 
   let n = ref 0 in
-  fun () -> incr n; id_of_string ("_" ^ (string_of_int !n))
+  fun () -> incr n; Names.id_of_string ("_" ^ (string_of_int !n))
 
 let add_anonymous_entry node =
   let id = anonymous_id () in
@@ -183,7 +216,7 @@ let add_absolutely_named_leaf sp obj =
   add_entry sp (Leaf obj)
 
 let add_leaf id obj =
-  if fst (current_prefix ()) = initial_path then 
+  if fst (current_prefix ()) = Names.initial_path then 
     error ("No session module started (use -top dir)");
   let oname = make_oname id in
   cache_object (oname,obj);
@@ -216,15 +249,28 @@ let add_frozen_state () =
 
 (* Modules. *)
 
+
 let is_something_opened = function 
     (_,OpenedSection _) -> true 
   | (_,OpenedModule _) -> true
   | (_,OpenedModtype _) -> true
   | _ -> false
 
+
+let current_mod_id () = 
+  try match find_entry_p is_something_opened with
+    | oname,OpenedModule (_,_,nametab) -> 
+	basename (fst oname)
+    | oname,OpenedModtype (_,nametab) -> 
+	basename (fst oname)
+    | _ -> error "you are not in a module"
+  with Not_found ->
+    error "no opened modules"
+
+
 let start_module export id mp nametab = 
   let dir = extend_dirpath (fst !path_prefix) id in
-  let prefix = dir,(mp,empty_dirpath) in
+  let prefix = dir,(mp,Names.empty_dirpath) in
   let oname = make_path id, make_kn id in
   if Nametab.exists_module dir then
     errorlabstrm "open_module" (pr_id id ++ str " already exists") ;
@@ -237,21 +283,31 @@ let end_module id =
   let oname,nametab = 
     try match find_entry_p is_something_opened with
       | oname,OpenedModule (_,_,nametab) -> 
-	  let sp = fst oname in
-	  let id' = basename sp in
-	  if id<>id' then error "this is not the last opened module"; 
-	  oname,nametab
-      | _,OpenedModtype _ ->
-	  error "there are some open module types"
-      | _,OpenedSection _ ->
-	  error "there are some open sections"
+	  let id' = basename (fst oname) in
+	    if id<>id' then 
+	      errorlabstrm "end_module" (str "last opened module is " ++ pr_id id'); 
+	    oname,nametab
+      | oname,OpenedModtype _ ->
+	  let id' = basename (fst oname) in
+	    errorlabstrm "end_module" 
+	      (str "module type " ++ pr_id id' ++ str " is still opened")
+      | oname,OpenedSection _ ->
+	  let id' = basename (fst oname) in
+	    errorlabstrm  "end_module" 
+	      (str "section " ++ pr_id id' ++ str " is still opened")
       | _ -> assert false
     with Not_found ->
       error "no opened modules"
   in
-  let (after,_,before) = split_lib oname in
+  let (after,modopening,before) = split_lib oname in
   lib_stk := before;
+  add_entry (make_oname id) (ClosedModule (List.rev_append after (List.rev modopening)));
   let prefix = !path_prefix in
+  (* LEM: This module business seems more complicated than sections;
+          shouldn't a backtrack into a closed module also do something
+          with global.ml, given that closing a module does?
+     TODO
+   *)
   recalc_path_prefix ();
   (* add_frozen_state must be called after processing the module, 
      because we cannot recache interactive modules  *) 
@@ -259,7 +315,7 @@ let end_module id =
 
 let start_modtype id mp nametab = 
   let dir = extend_dirpath (fst !path_prefix) id in
-  let prefix = dir,(mp,empty_dirpath) in
+  let prefix = dir,(mp,Names.empty_dirpath) in
   let sp = make_path id in
   let name = sp, make_kn id in
   if Nametab.exists_cci sp then
@@ -271,20 +327,26 @@ let start_modtype id mp nametab =
 let end_modtype id = 
   let sp,nametab = 
     try match find_entry_p is_something_opened with
-      | sp,OpenedModtype (_,nametab) -> 
-	  let id' = basename (fst sp) in
-	  if id<>id' then error "this is not the last opened module"; 
-	  sp,nametab
-      | _,OpenedModule _ ->
-	  error "there are some open modules"
-      | _,OpenedSection _ ->
-	  error "there are some open sections"
+      | oname,OpenedModtype (_,nametab) -> 
+	  let id' = basename (fst oname) in
+	  if id<>id' then 
+	    errorlabstrm "end_modtype" (str "last opened module type is " ++ pr_id id'); 
+	    oname,nametab
+      | oname,OpenedModule _ ->
+	  let id' = basename (fst oname) in
+	    errorlabstrm "end_modtype" 
+	      (str "module " ++ pr_id id' ++ str " is still opened")
+      | oname,OpenedSection _ ->
+	  let id' = basename (fst oname) in
+	    errorlabstrm "end_modtype" 
+	      (str "section " ++ pr_id id' ++ str " is still opened")
       | _ -> assert false
     with Not_found ->
       error "no opened module types"
   in
-  let (after,_,before) = split_lib sp in
+  let (after,modtypeopening,before) = split_lib sp in
   lib_stk := before;
+  add_entry (make_oname id) (ClosedModtype (List.rev_append after (List.rev modtypeopening)));
   let dir = !path_prefix in
   recalc_path_prefix ();
   (* add_frozen_state must be called after processing the module type.
@@ -310,9 +372,9 @@ let check_for_comp_unit () =
 let start_compilation s mp =
   if !comp_name <> None then
     error "compilation unit is already started";
-  if snd (snd (!path_prefix)) <> empty_dirpath then
+  if snd (snd (!path_prefix)) <> Names.empty_dirpath then
     error "some sections are already opened";
-  let prefix = s, (mp, empty_dirpath) in
+  let prefix = s, (mp, Names.empty_dirpath) in
   let _ = add_anonymous_entry (CompilingLibrary prefix) in
   comp_name := Some s;
   path_prefix := prefix
@@ -342,8 +404,8 @@ let end_compilation dir =
       | None -> anomaly "There should be a module name..."
       | Some m ->
 	  if m <> dir then anomaly 
-	    ("The current open module has name "^ (string_of_dirpath m) ^ 
-	       " and not " ^ (string_of_dirpath m));
+	    ("The current open module has name "^ (Names.string_of_dirpath m) ^ 
+	       " and not " ^ (Names.string_of_dirpath m));
   in
   let (after,_,before) = split_lib oname in
     comp_name := None;
@@ -383,65 +445,89 @@ let what_is_opened () = find_entry_p is_something_opened
    - the list of variables on which each inductive depends in this section
    - the list of substitution to do at section closing 
 *)
+type binding_kind = Explicit | Implicit
 
-type abstr_list = Sign.named_context Cmap.t * Sign.named_context KNmap.t
+type variable_info = Names.identifier * binding_kind * Term.constr option * Term.types
+type variable_context = variable_info list
+type abstr_list = variable_context Names.Cmap.t * variable_context Names.KNmap.t
 
 let sectab =
-  ref ([] : (identifier list * Cooking.work_list * abstr_list) list)
+  ref ([] : ((Names.identifier * binding_kind * Term.types option) list * Cooking.work_list * abstr_list) list)
 
 let add_section () =
-  sectab := ([],(Cmap.empty,KNmap.empty),(Cmap.empty,KNmap.empty)) :: !sectab
+  sectab := ([],(Names.Cmap.empty,Names.KNmap.empty),(Names.Cmap.empty,Names.KNmap.empty)) :: !sectab
 
-let add_section_variable id =
+let add_section_variable id impl keep =
   match !sectab with
     | [] -> () (* because (Co-)Fixpoint temporarily uses local vars *)
-    | (vars,repl,abs)::sl -> sectab := (id::vars,repl,abs)::sl
+    | (vars,repl,abs)::sl -> sectab := ((id,impl,keep)::vars,repl,abs)::sl
 
 let rec extract_hyps = function
-  | (id::idl,(id',_,_ as d)::hyps) when id=id' -> d :: extract_hyps (idl,hyps)
+  | ((id,impl,keep)::idl,(id',b,t)::hyps) when id=id' -> (id',impl,b,t) :: extract_hyps (idl,hyps)
+  | ((id,impl,Some ty)::idl,hyps) -> (id,impl,None,ty) :: extract_hyps (idl,hyps)
   | (id::idl,hyps) -> extract_hyps (idl,hyps)
   | [], _ -> []
 
+let instance_from_variable_context sign =
+  let rec inst_rec = function
+    | (id,b,None,_) :: sign -> id :: inst_rec sign
+    | _ :: sign -> inst_rec sign
+    | [] -> [] in
+  Array.of_list (inst_rec sign)
+
+let named_of_variable_context = List.map (fun (id,_,b,t) -> (id,b,t))
+
 let add_section_replacement f g hyps =
   match !sectab with
   | [] -> ()
   | (vars,exps,abs)::sl ->
     let sechyps = extract_hyps (vars,hyps) in
-    let args = Sign.instance_from_named_context (List.rev sechyps) in
-    sectab := (vars,f (Array.map Term.destVar args) exps,g sechyps abs)::sl
+    let args = instance_from_variable_context (List.rev sechyps) in
+    sectab := (vars,f args exps,g sechyps abs)::sl
 	
 let add_section_kn kn =
-  let f = (fun x (l1,l2) -> (l1,KNmap.add kn x l2)) in
+  let f x (l1,l2) = (l1,Names.KNmap.add kn x l2) in
   add_section_replacement f f
 
 let add_section_constant kn =
-  let f = (fun x (l1,l2) -> (Cmap.add kn x l1,l2)) in
+  let f x (l1,l2) = (Names.Cmap.add kn x l1,l2) in
   add_section_replacement f f
 
 let replacement_context () = pi2 (List.hd !sectab)
 
+let variables_context () = pi1 (List.hd !sectab)
+
 let section_segment_of_constant con =
-  Cmap.find con (fst (pi3 (List.hd !sectab)))
+  Names.Cmap.find con (fst (pi3 (List.hd !sectab)))
 
 let section_segment_of_mutual_inductive kn =
-  KNmap.find kn (snd (pi3 (List.hd !sectab)))
+  Names.KNmap.find kn (snd (pi3 (List.hd !sectab)))
+
+let rec list_mem_assoc_in_triple x = function
+    [] -> raise Not_found
+  | (a,_,_)::l -> compare a x = 0 or list_mem_assoc_in_triple x l
 
 let section_instance = function
-  | VarRef id -> [||]
+  | VarRef id ->
+      if list_mem_assoc_in_triple id (pi1 (List.hd !sectab)) then [||]
+      else raise Not_found
   | ConstRef con ->
-      Cmap.find con (fst (pi2 (List.hd !sectab)))
+      Names.Cmap.find con (fst (pi2 (List.hd !sectab)))
   | IndRef (kn,_) | ConstructRef ((kn,_),_) ->
-      KNmap.find kn (snd (pi2 (List.hd !sectab)))
+      Names.KNmap.find kn (snd (pi2 (List.hd !sectab)))
 
-let init () = sectab := []
-let freeze () = !sectab
-let unfreeze s = sectab := s
+let is_in_section ref =
+  try ignore (section_instance ref); true with Not_found -> false
+
+let init_sectab () = sectab := []
+let freeze_sectab () = !sectab
+let unfreeze_sectab s = sectab := s
 
 let _ = 
   Summary.declare_summary "section-context"
-    { Summary.freeze_function = freeze;
-      Summary.unfreeze_function = unfreeze;
-      Summary.init_function = init;
+    { Summary.freeze_function = freeze_sectab;
+      Summary.unfreeze_function = unfreeze_sectab;
+      Summary.init_function = init_sectab;
       Summary.survive_module = false;
       Summary.survive_section = false }
 
@@ -467,7 +553,7 @@ let open_section id =
       (*Pushed for the lifetime of the section: removed by unfrozing the summary*)
       Nametab.push_dir (Nametab.Until 1) dir (DirOpenSection prefix);
       path_prefix := prefix;
-      if !Options.xml_export then !xml_open_section id;
+      if !Flags.xml_export then !xml_open_section id;
       add_section ()
 
 
@@ -477,9 +563,9 @@ let open_section id =
 let discharge_item ((sp,_ as oname),e) =
   match e with
   | Leaf lobj ->
-      option_map (fun o -> (basename sp,o)) (discharge_object (oname,lobj))
+      Option.map (fun o -> (basename sp,o)) (discharge_object (oname,lobj))
   | FrozenState _ -> None
-  | ClosedSection -> None
+  | ClosedSection _ | ClosedModtype _ | ClosedModule _ -> None
   | OpenedSection _ | OpenedModtype _ | OpenedModule _ | CompilingLibrary _ ->
       anomaly "discharge_item"
 
@@ -487,28 +573,33 @@ let close_section id =
   let oname,fs = 
     try match find_entry_p is_something_opened with
       | oname,OpenedSection (_,fs) -> 
-	  if id <> basename (fst oname) then
-	    error "this is not the last opened section";
-	  (oname,fs)
+	  let id' = basename (fst oname) in 
+	    if id <> id' then 
+	      errorlabstrm "close_section" (str "Last opened section is " ++ pr_id id' ++ str ".");
+	    (oname,fs)
       | _ -> assert false 
     with Not_found ->
-      error "no opened section"
+      error "No opened section."
   in
-  let (secdecls,_,before) = split_lib oname in
+  let (secdecls,secopening,before) = split_lib oname in
   lib_stk := before;
   let full_olddir = fst !path_prefix in
   pop_path_prefix ();
-  add_entry (make_oname id) ClosedSection;
-  if !Options.xml_export then !xml_close_section id;
+  add_entry (make_oname id) (ClosedSection (List.rev_append secdecls (List.rev secopening)));
+  if !Flags.xml_export then !xml_close_section id;
   let newdecls = List.map discharge_item secdecls in
   Summary.section_unfreeze_summaries fs;
-  List.iter (option_iter (fun (id,o) -> add_discharged_leaf id o)) newdecls;
+  List.iter (Option.iter (fun (id,o) -> add_discharged_leaf id o)) newdecls;
   Cooking.clear_cooking_sharing ();
   Nametab.push_dir (Nametab.Until 1) full_olddir (DirClosedSection full_olddir)
 
 (*****************)
 (* Backtracking. *)
 
+let (inLabel,outLabel) =
+  declare_object {(default_object "DOT") with
+				classify_function = (fun _ -> Dispose)}
+
 let recache_decl = function
   | (sp, Leaf o) -> cache_object (sp,o)
   | (_,OpenedSection _) -> add_section ()
@@ -519,16 +610,58 @@ let recache_context ctx =
 
 let is_frozen_state = function (_,FrozenState _) -> true | _ -> false
 
-let reset_to sp =
-  let (_,_,before) = split_lib sp in
-  lib_stk := before;
+let has_top_frozen_state () =
+  let rec aux = function
+  | (sp, FrozenState _)::_ -> Some sp
+  | (sp, Leaf o)::t when object_tag o = "DOT" -> aux t
+  | _ -> None
+  in aux !lib_stk 
+
+let set_lib_stk new_lib_stk =
+  lib_stk := new_lib_stk;
   recalc_path_prefix ();
   let spf = match find_entry_p is_frozen_state with
     | (sp, FrozenState f) -> unfreeze_summaries f; sp
     | _ -> assert false
   in
   let (after,_,_) = split_lib spf in
-  recache_context after
+  try
+    recache_context after
+  with
+    | Not_found -> error "Tried to set environment to an incoherent state."
+
+let reset_to_gen test =
+  let (_,_,before) = split_lib_gen test in
+  set_lib_stk before
+
+let reset_to sp = reset_to_gen (fun x -> (fst x) = sp)
+
+let reset_to_state sp =
+  let (_,eq,before) = split_lib sp in
+  (* if eq a frozen state, we'll reset to it *)
+  match eq with
+  | [_,FrozenState f] -> lib_stk := eq@before;  unfreeze_summaries f
+  | _ -> error "Not a frozen state"
+
+
+(* LEM: TODO
+ * We will need to muck with frozen states in after, too!
+ * Not only FrozenState, but also those embedded in Opened(Section|Module|Modtype)
+ *)
+let delete_gen test =
+  let (after,equal,before) = split_lib_gen test in
+  let rec chop_at_dot = function
+    | [] as l -> l
+    | (_, Leaf o)::t when object_tag o = "DOT" -> t
+    | _::t -> chop_at_dot t
+  and chop_before_dot = function
+    | [] as l -> l
+    | (_, Leaf o)::t as l when object_tag o = "DOT" -> l
+    | _::t -> chop_before_dot t
+ in
+  set_lib_stk (List.rev_append (chop_at_dot after) (chop_before_dot before))
+
+let delete sp = delete_gen (fun x -> (fst x) = sp)
 
 let reset_name (loc,id) =
   let (sp,_) = 
@@ -539,10 +672,20 @@ let reset_name (loc,id) =
   in
   reset_to sp
 
+let remove_name (loc,id) =
+  let (sp,_) =
+    try
+      find_entry_p (fun (sp,_) -> let (_,spi) = repr_path (fst sp) in id = spi)
+    with Not_found ->
+      user_err_loc (loc,"remove_name",pr_id id ++ str ": no such entry")
+  in
+    delete sp
+
 let is_mod_node = function 
   | OpenedModule _ | OpenedModtype _ | OpenedSection _ 
-  | ClosedSection -> true 
-  | Leaf o -> let t = object_tag o in t = "MODULE" || t = "MODULE TYPE"
+  | ClosedModule _ | ClosedModtype _ | ClosedSection _  -> true 
+  | Leaf o -> let t = object_tag o in t = "MODULE" || t = "MODULE TYPE" 
+	|| t = "MODULE ALIAS"
   | _ -> false
 
 (* Reset on a module or section name in order to bypass constants with 
@@ -557,19 +700,7 @@ let reset_mod (loc,id) =
     with Not_found ->
       user_err_loc (loc,"reset_mod",pr_id id ++ str ": no such entry")
   in
-  lib_stk := before;
-  recalc_path_prefix ();
-  let spf = match find_entry_p is_frozen_state with
-    | (sp, FrozenState f) -> unfreeze_summaries f; sp
-    | _ -> assert false
-  in
-  let (after,_,_) = split_lib spf in
-  recache_context after
-
-
-let (inLabel,outLabel) =
-  declare_object {(default_object "DOT") with
-				classify_function = (fun _ -> Dispose)}
+  set_lib_stk before
 
 let mark_end_of_command, current_command_label, set_command_label =
   let n = ref 0 in
@@ -580,17 +711,23 @@ let mark_end_of_command, current_command_label, set_command_label =
   (fun () -> !n),
   (fun x -> n:=x)
 
-let rec reset_label_stk n stk =
-  match stk with
-      (sp,Leaf o)::tail when object_tag o = "DOT" && n = outLabel o -> sp
-    | _::tail -> reset_label_stk n tail
-    | [] -> error "Unknown state number"
+let is_label_n n x =
+  match x with
+    | (sp,Leaf o) when object_tag o = "DOT" && n = outLabel o -> true
+    | _ -> false
 
+(* Reset the label registered by [mark_end_of_command()] with number n. *)
 let reset_label n = 
-  let res = reset_label_stk n !lib_stk in
-  set_command_label (n-1); (* forget state numbers after n only if reset succeeded *)
-  reset_to res
-
+  let current = current_command_label() in
+  if n < current then
+    let res = reset_to_gen (is_label_n n) in
+    set_command_label (n-1); (* forget state numbers after n only if reset succeeded *)
+    res
+  else (* optimisation to avoid recaching when not necessary (why is it so long??) *)
+    match !lib_stk with
+      | [] -> ()
+      | x :: ls -> (lib_stk := ls;set_command_label (n-1))
+    
 let rec back_stk n stk =
   match stk with
       (sp,Leaf o)::tail when object_tag o = "DOT" ->
@@ -602,7 +739,7 @@ let back n = reset_to (back_stk n !lib_stk)
 
 (* State and initialization. *)
 
-type frozen = dir_path option * library_segment
+type frozen = Names.dir_path option * library_segment
 
 let freeze () = (!comp_name, !lib_stk)
 
@@ -646,16 +783,37 @@ let reset_initial () =
 let mp_of_global ref = 
   match ref with
     | VarRef id -> fst (current_prefix ())
-    | ConstRef cst -> con_modpath cst
-    | IndRef ind -> ind_modpath ind
-    | ConstructRef constr -> constr_modpath constr
+    | ConstRef cst -> Names.con_modpath cst
+    | IndRef ind -> Names.ind_modpath ind
+    | ConstructRef constr -> Names.constr_modpath constr
 
 let rec dp_of_mp modp =
   match modp with
-    | MPfile dp -> dp
-    | MPbound _ | MPself _ -> library_dp ()
-    | MPdot (mp,_) -> dp_of_mp mp
+    | Names.MPfile dp -> dp
+    | Names.MPbound _ | Names.MPself _ -> library_dp ()
+    | Names.MPdot (mp,_) -> dp_of_mp mp
 
+let rec split_mp mp = 
+  match mp with 
+    | Names.MPfile dp -> dp,  Names.empty_dirpath
+    | Names.MPdot (prfx, lbl) -> 
+	let mprec, dprec = split_mp prfx in 
+	  mprec, Names.make_dirpath (Names.id_of_string (Names.string_of_label lbl) :: (Names.repr_dirpath dprec))
+    | Names.MPself msid -> let (_, id, dp) = Names.repr_msid msid in  library_dp(), Names.make_dirpath [Names.id_of_string id]
+    | Names.MPbound mbid -> let (_, id, dp) = Names.repr_mbid mbid in  library_dp(), Names.make_dirpath [Names.id_of_string id]
+
+let split_modpath mp =
+  let rec aux = function
+    | Names.MPfile dp -> dp, []
+    | Names.MPbound mbid -> 
+	library_dp (), [Names.id_of_mbid mbid]
+    | Names.MPself msid -> library_dp (), [Names.id_of_msid msid]
+    | Names.MPdot (mp,l) -> let (mp', lab) = aux mp in
+			      (mp', Names.id_of_label l :: lab)
+  in 
+  let (mp, l) = aux mp in
+    mp, l
+			
 let library_part ref =
   match ref with 
     | VarRef id -> library_dp ()
@@ -678,13 +836,21 @@ let remove_section_part ref =
 (************************)
 (* Discharging names *)
 
+let pop_kn kn =
+  let (mp,dir,l) = Names.repr_kn kn in
+  Names.make_kn mp (dirpath_prefix dir) l
+
+let pop_con con = 
+  let (mp,dir,l) = Names.repr_con con in
+  Names.make_con mp (dirpath_prefix dir) l
+
 let con_defined_in_sec kn = 
-  let _,dir,_ = repr_con kn in
-  dir <> empty_dirpath && fst (split_dirpath dir) = snd (current_prefix ())
+  let _,dir,_ = Names.repr_con kn in
+  dir <> Names.empty_dirpath && fst (split_dirpath dir) = snd (current_prefix ())
 
 let defined_in_sec kn = 
-  let _,dir,_ = repr_kn kn in
-  dir <> empty_dirpath && fst (split_dirpath dir) = snd (current_prefix ())
+  let _,dir,_ = Names.repr_kn kn in
+  dir <> Names.empty_dirpath && fst (split_dirpath dir) = snd (current_prefix ())
 
 let discharge_global = function
   | ConstRef kn when con_defined_in_sec kn ->
