@@ -6,7 +6,7 @@
 (*         *       GNU Lesser General Public License Version 2.1        *)
 (************************************************************************)
 
-(* $Id: ppvernac.ml 9562 2007-01-31 09:00:36Z msozeau $ *)  
+(* $Id: ppvernac.ml 11809 2009-01-20 11:39:55Z aspiwack $ *)  
 
 open Pp
 open Names
@@ -28,7 +28,7 @@ open Tacinterp
 
 let pr_spc_lconstr = pr_sep_com spc pr_lconstr_expr
 
-let pr_lident (b,_ as loc,id) =
+let pr_lident (loc,id) =
   if loc <> dummy_loc then
     let (b,_) = unloc loc in
     pr_located pr_id (make_loc (b,b+String.length(string_of_id id)),id)
@@ -39,7 +39,7 @@ let string_of_fqid fqid =
 
 let pr_fqid fqid = str (string_of_fqid fqid)
 
-let pr_lfqid (_,_ as loc,fqid) =
+let pr_lfqid (loc,fqid) =
   if loc <> dummy_loc then
    let (b,_) = unloc loc in
     pr_located pr_fqid (make_loc (b,b+String.length(string_of_fqid fqid)),fqid)
@@ -50,7 +50,7 @@ let pr_lname = function
     (loc,Name id) -> pr_lident (loc,id)
   | lna -> pr_located pr_name lna
 
-let pr_ltac_id = Nameops.pr_id
+let pr_ltac_id = Libnames.pr_reference
 
 let pr_module = Libnames.pr_reference
 
@@ -133,9 +133,11 @@ let pr_in_out_modules = function
   | SearchOutside [] -> mt()
   | SearchOutside l -> spc() ++ str"outside" ++ spc() ++ prlist_with_sep sep pr_module l
 
-let pr_search_about = function
-  | SearchRef r -> pr_reference r
-  | SearchString s -> qs s
+let pr_search_about (b,c) = 
+  (if b then str "-" else mt()) ++
+  match c with
+  | SearchSubPattern p -> pr_constr_pattern_expr p
+  | SearchString (s,sc) -> qs s ++ pr_opt (fun sc -> str "%" ++ str sc) sc
 
 let pr_search a b pr_p = match a with
   | SearchHead qid -> str"Search" ++ spc() ++ pr_reference qid ++ pr_in_out_modules b
@@ -144,10 +146,14 @@ let pr_search a b pr_p = match a with
   | SearchAbout sl -> str"SearchAbout" ++ spc() ++ str "[" ++ prlist_with_sep spc pr_search_about sl ++ str "]" ++ pr_in_out_modules b
 
 let pr_locality local = if local then str "Local " else str ""
+let pr_non_locality local = if local then str "" else str "Global "
 
-let pr_explanation imps = function
-  | ExplByPos n -> pr_id (Impargs.name_of_implicit (List.nth imps (n-1)))
-  | ExplByName id -> pr_id id
+let pr_explanation (e,b,f) =
+  let a = match e with
+  | ExplByPos (n,_) -> anomaly "No more supported"
+  | ExplByName id -> pr_id id in
+  let a = if f then str"!" ++ a else a in
+    if b then str "[" ++ a ++ str "]" else a
 
 let pr_class_rawexpr = function
   | FunClass -> str"Funclass"
@@ -161,6 +167,9 @@ let pr_option_ref_value = function
 let pr_printoption a b = match a with
   | Goptions.PrimaryTable table -> str table ++ pr_opt (prlist_with_sep sep pr_option_ref_value) b
   | Goptions.SecondaryTable (table,field) -> str table ++ spc() ++ str field ++ pr_opt (prlist_with_sep sep pr_option_ref_value) b
+  | Goptions.TertiaryTable (table,field1,field2) -> str table ++ spc() ++
+      str field1 ++ spc() ++ str field2 ++ 
+      pr_opt (prlist_with_sep sep pr_option_ref_value) b
 
 let pr_set_option a b = 
   let pr_opt_value = function 
@@ -184,16 +193,23 @@ let pr_hints local db h pr_c pr_pat =
   let pph =
     match h with
     | HintsResolve l ->
-        str "Resolve " ++ prlist_with_sep sep pr_c l
+        str "Resolve " ++ prlist_with_sep sep 
+	  (fun (pri, _, c) -> pr_c c ++ 
+	    match pri with Some x -> spc () ++ str"(" ++ int x ++ str")" | None -> mt ())
+	  l
     | HintsImmediate l ->
         str"Immediate" ++ spc() ++ prlist_with_sep sep pr_c l
     | HintsUnfold l ->
         str "Unfold " ++ prlist_with_sep sep pr_reference l
+    | HintsTransparency (l, b) ->
+        str (if b then "Transparent " else "Opaque ") ++ prlist_with_sep sep 
+	  pr_reference l
     | HintsConstructors c ->
         str"Constructors" ++ spc() ++ prlist_with_sep spc pr_reference c
-    | HintsExtern (n,c,tac) ->
-        str "Extern" ++ spc() ++ int n ++ spc() ++ pr_pat c ++ str" =>" ++
-        spc() ++ pr_raw_tactic tac
+    | HintsExtern (n,c,tac) -> 
+	let pat = match c with None -> mt () | Some pat -> pr_pat pat in
+          str "Extern" ++ spc() ++ int n ++ spc() ++ pat ++ str" =>" ++
+          spc() ++ pr_raw_tactic tac
     | HintsDestruct(name,i,loc,c,tac) ->
         str "Destruct " ++ pr_id name ++ str" :=" ++ spc() ++
         hov 0 (int i ++ spc() ++ pr_destruct_location loc ++ spc() ++
@@ -215,6 +231,18 @@ let rec pr_module_type pr_c = function
       let m = pr_module_type pr_c mty in
       let p = pr_with_declaration pr_c decl in
       m ++ spc() ++ str"with" ++ spc() ++ p
+  | CMTEapply (fexpr,mexpr)->
+      let f = pr_module_type pr_c fexpr in
+      let m = pr_module_expr mexpr in
+	f ++ spc () ++ m 
+	  
+and  pr_module_expr = function
+  | CMEident qid -> pr_located pr_qualid qid
+  | CMEapply (me1,(CMEident _ as me2)) ->
+      pr_module_expr me1 ++ spc() ++ pr_module_expr me2
+  | CMEapply (me1,me2) ->
+      pr_module_expr me1 ++ spc() ++
+      hov 1 (str"(" ++ pr_module_expr me2 ++ str")")
 
 let pr_of_module_type prc (mty,b) =
   str (if b then ":" else "<:") ++
@@ -247,16 +275,8 @@ let pr_module_binders l pr_c =
 
 let pr_module_binders_list l pr_c = pr_module_binders l pr_c
 
-let rec pr_module_expr = function
-  | CMEident qid -> pr_located pr_qualid qid
-  | CMEapply (me1,(CMEident _ as me2)) ->
-      pr_module_expr me1 ++ spc() ++ pr_module_expr me2
-  | CMEapply (me1,me2) ->
-      pr_module_expr me1 ++ spc() ++
-      hov 1 (str"(" ++ pr_module_expr me2 ++ str")")
-
 let pr_type_option pr_c = function
-  | CHole loc -> mt()
+  | CHole (loc, k) -> mt()
   | _ as c -> brk(0,2) ++ str":" ++ pr_c c
 
 let pr_decl_notation prc =
@@ -273,11 +293,23 @@ let pr_binders_arg =
 let pr_and_type_binders_arg bl =
   pr_binders_arg bl
 
-let pr_onescheme (id,dep,ind,s) =
-  hov 0 (pr_lident id ++ str" :=") ++ spc() ++
-  hov 0 ((if dep then str"Induction for" else str"Minimality for")
-  ++ spc() ++ pr_reference ind) ++ spc() ++ 
-  hov 0 (str"Sort" ++ spc() ++ pr_rawsort s)
+let pr_onescheme (idop,schem) =
+  match schem with 
+  | InductionScheme (dep,ind,s) ->
+    (match idop with
+      | Some id -> hov 0 (pr_lident id ++ str" :=") ++ spc()
+      | None -> spc ()
+    ) ++
+    hov 0 ((if dep then str"Induction for" else str"Minimality for")
+    ++ spc() ++ pr_reference ind) ++ spc() ++ 
+    hov 0 (str"Sort" ++ spc() ++ pr_rawsort s)
+  | EqualityScheme ind -> 
+    (match idop with
+      | Some id -> hov 0 (pr_lident id ++ str" :=") ++ spc()
+      | None -> spc()
+    ) ++
+    hov 0 (str"Equality for")
+    ++ spc() ++ pr_reference ind
 
 let begin_of_inductive = function
     [] -> 0
@@ -299,7 +331,7 @@ let pr_assumption_token many = function
       str (if many then "Parameters" else "Parameter")
   | (Global,Conjectural) -> str"Conjecture"
   | (Local,Conjectural) -> 
-      anomaly "Don't know how to translate a local conjecture"
+      anomaly "Don't know how to beautify a local conjecture"
 
 let pr_params pr_c (xl,(c,t)) =
   hov 2 (prlist_with_sep sep pr_lident xl ++ spc() ++
@@ -376,6 +408,27 @@ let make_pr_vernac pr_constr pr_lconstr =
 let pr_constrarg c = spc () ++ pr_constr c in
 let pr_lconstrarg c = spc () ++ pr_lconstr c in
 let pr_intarg n = spc () ++ int n in
+(* let pr_lident_constr sep (i,c) = pr_lident i ++ sep ++ pr_constrarg c in *)
+let pr_record_field (x, ntn) = 
+  let prx = match x with
+    | (oc,AssumExpr (id,t)) ->
+	hov 1 (pr_lname id ++
+		  (if oc then str" :>" else str" :") ++ spc() ++
+		  pr_lconstr_expr t)
+    | (oc,DefExpr(id,b,opt)) -> (match opt with
+      | Some t ->
+          hov 1 (pr_lname id ++
+                    (if oc then str" :>" else str" :") ++ spc() ++
+                    pr_lconstr_expr t ++ str" :=" ++ pr_lconstr b)
+      | None ->
+          hov 1 (pr_lname id ++ str" :=" ++ spc() ++
+                    pr_lconstr b)) in
+    prx ++ pr_decl_notation pr_constr ntn
+in
+let pr_record_decl b c fs = 
+    pr_opt pr_lident c ++ str"{"  ++
+    hv 0 (prlist_with_sep pr_semicolon pr_record_field fs ++ str"}")
+in
 
 let rec pr_vernac = function
   
@@ -388,6 +441,7 @@ let rec pr_vernac = function
   | VernacAbort id -> str"Abort" ++ pr_opt pr_lident id
   | VernacResume id -> str"Resume" ++ pr_opt pr_lident id
   | VernacUndo i -> if i=1 then str"Undo" else str"Undo" ++ pr_intarg i
+  | VernacUndoTo i -> str"Undo" ++ spc() ++ str"To" ++ pr_intarg i
   | VernacBacktrack (i,j,k) -> 
       str "Backtrack" ++  spc() ++ prlist_with_sep sep int [i;j;k]
   | VernacFocus i -> str"Focus" ++ pr_opt int i
@@ -417,6 +471,7 @@ let rec pr_vernac = function
   | VernacCheckGuard -> str"Guarded"
 
   (* Resetting *)
+  | VernacRemoveName id -> str"Remove" ++ spc() ++ pr_lident id
   | VernacResetName id -> str"Reset" ++ spc() ++ pr_lident id
   | VernacResetInitial -> str"Reset Initial"
   | VernacBack i -> if i=1 then str"Back" else str"Back" ++ pr_intarg i
@@ -434,7 +489,6 @@ let rec pr_vernac = function
   | VernacLoad (f,s) -> str"Load" ++ if f then (spc() ++ str"Verbose"
   ++ spc()) else spc()  ++ qs s
   | VernacTime v -> str"Time" ++ spc() ++ pr_vernac v
-  | VernacVar id -> pr_lident id
   
   (* Syntax *) 
   | VernacTacticNotation (n,r,e) -> pr_grammar_tactic_rule n ("",r,e)
@@ -450,7 +504,8 @@ let rec pr_vernac = function
   | VernacArgumentsScope (local,q,scl) -> let pr_opt_scope = function 
       |	None -> str"_"
       |	Some sc -> str sc in 
-    str"Arguments Scope" ++ spc() ++ pr_reference q ++ spc() ++ str"[" ++ prlist_with_sep sep pr_opt_scope scl ++ str"]"
+    str"Arguments Scope" ++ spc() ++ pr_non_locality local ++ pr_reference q 
+    ++ spc() ++ str"[" ++ prlist_with_sep sep pr_opt_scope scl ++ str"]"
   | VernacInfix (local,(s,mv),q,sn) -> (* A Verifier *)
       hov 0 (hov 0 (str"Infix " ++ pr_locality local
       ++ qs s ++ str " :=" ++ spc() ++ pr_reference q) ++
@@ -482,7 +537,7 @@ let rec pr_vernac = function
         | None -> mt()
         | Some r ->
             str"Eval" ++ spc() ++
-            pr_red_expr (pr_constr, pr_lconstr, pr_reference) r ++
+            pr_red_expr (pr_constr, pr_lconstr, pr_or_by_notation pr_reference) r ++
             str" in" ++ spc() in
       let pr_def_body = function
         | DefineBody (bl,red,body,d) ->
@@ -499,12 +554,15 @@ let rec pr_vernac = function
         | None -> mt()
         | Some cc -> str" :=" ++ spc() ++ cc))
 
-  | VernacStartTheoremProof (ki,id,(bl,c),b,d) ->
-      hov 1 (pr_thm_token ki ++ spc() ++ pr_lident id ++ spc() ++
-      (match bl with
-        | [] -> mt()
-        | _ -> pr_binders bl ++ spc())
-      ++ str":" ++ pr_spc_lconstr c)
+  | VernacStartTheoremProof (ki,l,_,_) ->
+      let pr_statement head (id,(bl,c)) =
+        hov 0
+          (head ++ pr_opt pr_lident id ++ spc() ++
+              (match bl with [] -> mt() | _ -> pr_binders bl ++ spc()) ++
+              str":" ++ pr_spc_lconstr c) in
+      hov 1 (pr_statement (pr_thm_token ki) (List.hd l) ++
+             prlist (pr_statement (spc () ++ str "with")) (List.tl l))
+
   | VernacEndProof Admitted -> str"Admitted"
   | VernacEndProof (Proved (opac,o)) -> (match o with
     | None -> if opac then str"Qed" else str"Defined"
@@ -513,7 +571,7 @@ let rec pr_vernac = function
       |	Some tok -> str"Save" ++ spc() ++ pr_thm_token tok ++ spc() ++ pr_lident id)) 
   | VernacExactProof c ->
       hov 2 (str"Proof" ++ pr_lconstrarg c)
-  | VernacAssumption (stre,l) ->
+  | VernacAssumption (stre,_,l) ->
       let n = List.length (List.flatten (List.map fst (List.map snd l))) in
       hov 2
         (pr_assumption_token (n > 1) stre ++ spc() ++ 
@@ -524,57 +582,61 @@ let rec pr_vernac = function
         hov 2 (pr_lident id ++ str" " ++
                (if coe then str":>" else str":") ++
                 pr_spc_lconstr c) in
-      let pr_constructor_list l = match l with
-        | [] -> mt()
-        | _ ->
+      let pr_constructor_list b l = match l with
+        | Constructors [] -> mt()
+        | Constructors l ->
             pr_com_at (begin_of_inductive l) ++
             fnl() ++
             str (if List.length l = 1 then "   " else " | ") ++
-            prlist_with_sep (fun _ -> fnl() ++ str" | ") pr_constructor l in
-      let pr_oneind key ((id,indpar,s,lc),ntn) =
-	hov 0 (
-          str key ++ spc() ++
-          pr_lident id ++ pr_and_type_binders_arg indpar ++ spc() ++ str":" ++ 
-	  spc() ++ pr_lconstr_expr s ++ 
-	  str" :=") ++ pr_constructor_list lc ++ 
-	pr_decl_notation pr_constr ntn in
-
-      hov 1 (pr_oneind (if f then "Inductive" else "CoInductive") (List.hd l))
+            prlist_with_sep (fun _ -> fnl() ++ str" | ") pr_constructor l 
+       | RecordDecl (c,fs) ->   
+	    spc() ++
+	    pr_record_decl b c fs in
+      let pr_oneind key (((coe,id),indpar,s,k,lc),ntn) =
+	let kw =
+	  str (match k with Record -> "Record" | Structure -> "Structure"  
+		 | Inductive_kw -> "Inductive" | CoInductive -> "CoInductive"
+		 | Class b -> if b then "Definitional Class" else "Class")
+	in
+	  hov 0 (
+	    kw ++ spc() ++
+              (if coe then str" > " else str" ") ++ pr_lident id ++
+              pr_and_type_binders_arg indpar ++ spc() ++ 
+	      Option.cata (fun s -> str":" ++ spc() ++ pr_lconstr_expr s) (mt()) s ++ 
+	      str" :=") ++ pr_constructor_list k lc ++ 
+	    pr_decl_notation pr_constr ntn 
+      in
+      hov 1 (pr_oneind (if (Decl_kinds.recursivity_flag_of_kind f) then "Inductive" else "CoInductive") (List.hd l))
       ++ 
       (prlist (fun ind -> fnl() ++ hov 1 (pr_oneind "with" ind)) (List.tl l))
 
 
   | VernacFixpoint (recs,b) ->
       let name_of_binder = function
-        | LocalRawAssum (nal,_) -> nal
+        | LocalRawAssum (nal,_,_) -> nal
         | LocalRawDef (_,_) -> [] in
       let pr_onerec = function
-        | (id,(n,ro),bl,type_,def),ntn ->
+        | ((loc,id),(n,ro),bl,type_,def),ntn ->
             let (bl',def,type_) =
-              if Options.do_translate() then extract_def_binders def type_
+              if Flags.do_beautify() then extract_def_binders def type_
               else ([],def,type_) in
             let bl = bl @ bl' in
             let ids = List.flatten (List.map name_of_binder bl) in
             let annot =
 	      match n with 
 		| None -> mt () 
-		| Some n -> 
-		    let name =
-		      try snd (List.nth ids n)
-		      with Failure _ ->
-			warn (str "non-printable fixpoint \""++pr_id id++str"\"");
-			Anonymous in
+		| Some (loc, id) -> 
 		    match (ro : Topconstr.recursion_order_expr) with
 			CStructRec -> 
 			  if List.length ids > 1 then 
-			    spc() ++ str "{struct " ++ pr_name name ++ str"}"
+			    spc() ++ str "{struct " ++ pr_id id ++ str"}"
 			  else mt()
 		      | CWfRec c -> 
 			  spc() ++ str "{wf " ++ pr_lconstr_expr c ++ spc() ++ 
-			    pr_name name ++ str"}"
+			    pr_id id ++ str"}"
 		      | CMeasureRec c -> 
 			  spc() ++ str "{measure " ++ pr_lconstr_expr c ++ spc() ++ 
-			    pr_name name ++ str"}"
+			    pr_id id ++ str"}"
 	    in
             pr_id id ++ pr_binders_arg bl ++ annot ++ spc()
             ++ pr_type_option (fun c -> spc() ++ pr_lconstr_expr c) type_
@@ -586,9 +648,9 @@ let rec pr_vernac = function
         prlist_with_sep (fun _ -> fnl() ++ fnl() ++ str"with ") pr_onerec recs)
 
   | VernacCoFixpoint (corecs,b) ->
-      let pr_onecorec ((id,bl,c,def),ntn) =
+      let pr_onecorec (((loc,id),bl,c,def),ntn) =
         let (bl',def,c) =
-              if Options.do_translate() then extract_def_binders def c
+              if Flags.do_beautify() then extract_def_binders def c
               else ([],def,c) in
         let bl = bl @ bl' in
         pr_id id ++ spc() ++ pr_binders bl ++ spc() ++ str":" ++
@@ -602,32 +664,13 @@ let rec pr_vernac = function
   | VernacScheme l ->
       hov 2 (str"Scheme" ++ spc() ++
              prlist_with_sep (fun _ -> fnl() ++ str"with ") pr_onescheme l)
+  | VernacCombinedScheme (id, l) ->
+      hov 2 (str"Combined Scheme" ++ spc() ++
+	       pr_lident id ++ spc() ++ str"from" ++ spc() ++
+	       prlist_with_sep (fun _ -> fnl() ++ str", ") pr_lident l)
+	
 
   (* Gallina extensions *)
-  | VernacRecord (b,(oc,name),ps,s,c,fs) ->
-      let pr_record_field = function
-        | (oc,AssumExpr (id,t)) ->
-            hov 1 (pr_lname id ++
-            (if oc then str" :>" else str" :") ++ spc() ++
-            pr_lconstr_expr t)
-        | (oc,DefExpr(id,b,opt)) -> (match opt with
-	    | Some t ->
-                hov 1 (pr_lname id ++
-                (if oc then str" :>" else str" :") ++ spc() ++
-                pr_lconstr_expr t ++ str" :=" ++ pr_lconstr b)
-	    | None ->
-                hov 1 (pr_lname id ++ str" :=" ++ spc() ++
-                pr_lconstr b)) in
-      hov 2
-        (str (if b then "Record" else "Structure") ++
-         (if oc then str" > " else str" ") ++ pr_lident name ++ 
-          pr_and_type_binders_arg ps ++ str" :" ++ spc() ++ 
-	  pr_lconstr_expr s ++ str" := " ++
-         (match c with
-           | None -> mt()
-           | Some sc -> pr_lident sc) ++
-	spc() ++ str"{"  ++
-        hv 0 (prlist_with_sep pr_semicolon pr_record_field fs ++ str"}"))
   | VernacBeginSection id -> hov 2 (str"Section" ++ spc () ++ pr_lident id)
   | VernacEndSegment id -> hov 2 (str"End" ++ spc() ++ pr_lident id)
   | VernacRequire (exp,spe,l) -> hov 2
@@ -654,24 +697,53 @@ let rec pr_vernac = function
 	  str"Local" ++ spc() | Global -> spc()) ++ pr_lident id ++ 
 	spc() ++ str":" ++ spc() ++ pr_class_rawexpr c1 ++ spc() ++ str">->" ++
 	spc() ++ pr_class_rawexpr c2)
+	
+ | VernacInstance (glob, sup, (instid, bk, cl), props, pri) -> 
+     hov 1 (
+       pr_non_locality (not glob) ++
+       str"Instance" ++ spc () ++ 
+	 pr_and_type_binders_arg sup ++
+	 str"=>" ++ spc () ++ 
+	 (match snd instid with Name id -> pr_lident (fst instid, id) ++ spc () ++ str":" ++ spc () | Anonymous -> mt ()) ++
+	 pr_constr_expr cl ++ spc () ++
+	 spc () ++ str":=" ++ spc () ++
+	 pr_constr_expr props)
 
+ | VernacContext l ->
+     hov 1 (
+       str"Context" ++ spc () ++ str"[" ++ spc () ++
+	 pr_and_type_binders_arg l ++ spc () ++ str "]")
+	      
+
+ | VernacDeclareInstance id ->
+     hov 1 (str"Instance" ++ spc () ++ pr_lident id)
+       
   (* Modules and Module Types *)
   | VernacDefineModule (export,m,bl,ty,bd) ->
       let b = pr_module_binders_list bl pr_lconstr in 
       hov 2 (str"Module" ++ spc() ++ pr_require_token export ++
-             pr_lident m ++ b ++
-             pr_opt (pr_of_module_type pr_lconstr) ty ++
-             pr_opt (fun me -> str ":= " ++ pr_module_expr me) bd)
+               pr_lident m ++ b ++
+               pr_opt (pr_of_module_type pr_lconstr) ty ++
+               pr_opt (fun me -> str ":= " ++ pr_module_expr me) bd)
   | VernacDeclareModule (export,id,bl,m1) ->
       let b = pr_module_binders_list bl pr_lconstr in 
-      hov 2 (str"Declare Module" ++ spc() ++ pr_require_token export ++
+	hov 2 (str"Declare Module" ++ spc() ++ pr_require_token export ++
              pr_lident id ++ b ++
              pr_of_module_type pr_lconstr m1)
   | VernacDeclareModuleType (id,bl,m) ->
       let b = pr_module_binders_list bl pr_lconstr in 
-      hov 2 (str"Module Type " ++ pr_lident id ++ b ++
-             pr_opt (fun mt -> str ":= " ++ pr_module_type pr_lconstr mt) m)
-
+	hov 2 (str"Module Type " ++ pr_lident id ++ b ++
+		 pr_opt (fun mt -> str ":= " ++ pr_module_type pr_lconstr mt) m)
+  | VernacInclude (in_ast) ->
+      begin
+	match in_ast with
+	  | CIMTE mty ->
+	      hov 2 (str"Include" ++ 
+		       (fun mt -> str " " ++ pr_module_type pr_lconstr mt) mty)
+	  | CIME mexpr ->
+	      hov 2 (str"Include" ++ 
+		       (fun me -> str " " ++ pr_module_expr me) mexpr)
+      end
   (* Solving *)
   | VernacSolve (i,tac,deftac) ->
       (if i = 1 then mt() else int i ++ str ": ") ++
@@ -714,49 +786,61 @@ let rec pr_vernac = function
 
   (* Commands *)
   | VernacDeclareTacticDefinition (rc,l) ->
-      let pr_tac_body (id, body) =
+      let pr_tac_body (id, redef, body) =
         let idl, body =
           match body with
 	    | Tacexpr.TacFun (idl,b) -> idl,b
             | _ -> [], body in
-        pr_located pr_ltac_id id ++ 
+        pr_ltac_id id ++ 
 	prlist (function None -> str " _"
                        | Some id -> spc () ++ pr_id id) idl
-	++ str" :=" ++ brk(1,1) ++
-	let idl = List.map out_some (List.filter (fun x -> not (x=None)) idl)in
+	++ (if redef then str" ::=" else str" :=") ++ brk(1,1) ++
+	let idl = List.map Option.get (List.filter (fun x -> not (x=None)) idl)in
         pr_raw_tactic_env 
-	  (idl @ List.map snd (List.map fst l))
+	  (idl @ List.map coerce_global_to_id 
+	    (List.map (fun (x, _, _) -> x) (List.filter (fun (_, redef, _) -> not redef) l)))
 	  (Global.env())
 	  body in
       hov 1
-        (((*if !Options.p1 then
-	  (if rc then str "Recursive " else mt()) ++
-	  str "Tactic Definition " else*)
-	    (* Rec by default *) str "Ltac ") ++
+        ((str "Ltac ") ++
         prlist_with_sep (fun () -> fnl() ++ str"with ") pr_tac_body l)
+  | VernacCreateHintDb (local,dbname,b) ->
+      hov 1 (str "Create " ++ pr_locality local ++ str "HintDb " ++ str dbname ++ (if b then str" discriminated" else mt ()))
   | VernacHints (local,dbnames,h) ->
-      pr_hints local dbnames h pr_constr pr_pattern_expr
-  | VernacSyntacticDefinition (id,c,local,onlyparsing) ->
+      pr_hints local dbnames h pr_constr pr_constr_pattern_expr
+  | VernacSyntacticDefinition (id,(ids,c),local,onlyparsing) ->
       hov 2
-        (str"Notation " ++ pr_locality local ++ pr_id id ++ str" :=" ++
-         pr_constrarg c ++
+        (str"Notation " ++ pr_locality local ++ pr_lident id ++ 
+	 prlist_with_sep spc pr_id ids ++ str" :=" ++ pr_constrarg c ++
          pr_syntax_modifiers (if onlyparsing then [SetOnlyParsing] else []))
   | VernacDeclareImplicits (local,q,None) ->
       hov 2 (str"Implicit Arguments" ++ spc() ++ pr_reference q)
-  | VernacDeclareImplicits (local,q,Some l) ->
-      let r = Nametab.global q in
-      Impargs.declare_manual_implicits local r l;
-      let imps = Impargs.implicits_of_global r in
-      hov 1 (str"Implicit Arguments" ++ spc() ++ pr_reference q ++ spc() ++
-             str"[" ++ prlist_with_sep sep (pr_explanation imps) l ++ str"]")
+  | VernacDeclareImplicits (local,q,Some imps) ->
+      hov 1 (str"Implicit Arguments " ++ pr_non_locality local ++
+	spc() ++ pr_reference q ++ spc() ++
+	str"[" ++ prlist_with_sep sep pr_explanation imps ++ str"]")
   | VernacReserve (idl,c) ->
       hov 1 (str"Implicit Type" ++
         str (if List.length idl > 1 then "s " else " ") ++
         prlist_with_sep spc pr_lident idl ++ str " :" ++ spc () ++
         pr_lconstr c)
-  | VernacSetOpacity (fl,l) ->
-      hov 1 ((if fl then str"Opaque" else str"Transparent") ++
+  | VernacSetOpacity(b,[k,l]) when k=Conv_oracle.transparent ->
+      hov 1 (str"Transparent" ++ pr_non_locality b ++
              spc() ++ prlist_with_sep sep pr_reference l)
+  | VernacSetOpacity(b,[Conv_oracle.Opaque,l]) ->
+      hov 1 (str"Opaque" ++ pr_non_locality b ++
+             spc() ++ prlist_with_sep sep pr_reference l)
+  | VernacSetOpacity (local,l) ->
+      let pr_lev = function
+          Conv_oracle.Opaque -> str"opaque"
+        | Conv_oracle.Expand -> str"expand"
+        | l when l = Conv_oracle.transparent -> str"transparent"
+        | Conv_oracle.Level n -> int n in
+      let pr_line (l,q) =
+        hov 2 (pr_lev l ++ spc() ++
+               str"[" ++ prlist_with_sep sep pr_reference q ++ str"]") in
+      hov 1 (pr_locality local ++ str"Strategy" ++ spc() ++
+             hv 0 (prlist_with_sep sep pr_line l))
   | VernacUnsetOption na ->
       hov 1 (str"Unset" ++ spc() ++ pr_printoption na None)
   | VernacSetOption (na,v) -> hov 2 (str"Set" ++ spc() ++ pr_set_option na v)
@@ -768,11 +852,11 @@ let rec pr_vernac = function
       let pr_mayeval r c = match r with 
       | Some r0 ->
           hov 2 (str"Eval" ++ spc() ++
-          pr_red_expr (pr_constr,pr_lconstr,pr_reference) r0 ++
+          pr_red_expr (pr_constr,pr_lconstr,pr_or_by_notation pr_reference) r0 ++
           spc() ++ str"in" ++ spc () ++ pr_constr c)
       | None -> hov 2 (str"Check" ++ spc() ++ pr_constr c) 
       in 
-      (if io = None then mt() else int (out_some io) ++ str ": ") ++ 
+      (if io = None then mt() else int (Option.get io) ++ str ": ") ++ 
       pr_mayeval r c
   | VernacGlobalCheck c -> hov 2 (str"Type" ++ pr_constrarg c)
   | VernacPrint p -> 
@@ -780,15 +864,16 @@ let rec pr_vernac = function
 	| PrintFullContext -> str"Print All"
 	| PrintSectionContext s ->
             str"Print Section" ++ spc() ++ Libnames.pr_reference s
-	| PrintGrammar (uni,ent) ->
-            msgerrnl (str "warning: no direct translation of Print Grammar entry"); 
+	| PrintGrammar ent ->
             str"Print Grammar" ++ spc() ++ str ent
-	| PrintLoadPath -> str"Print LoadPath"
+	| PrintLoadPath dir -> str"Print LoadPath" ++ pr_opt pr_dirpath dir
 	| PrintModules -> str"Print Modules"
 	| PrintMLLoadPath -> str"Print ML Path"
 	| PrintMLModules -> str"Print ML Modules"
 	| PrintGraph -> str"Print Graph"
 	| PrintClasses -> str"Print Classes"
+	| PrintTypeClasses -> str"Print TypeClasses"
+	| PrintInstances qid -> str"Print Instances" ++ spc () ++ pr_reference qid
 	| PrintLtac qid -> str"Print Ltac" ++ spc() ++ pr_reference qid
 	| PrintCoercions -> str"Print Coercions"
 	| PrintCoercionPaths (s,t) -> str"Print Coercion Paths" ++ spc() ++ pr_class_rawexpr s ++ spc() ++ pr_class_rawexpr t
@@ -805,14 +890,16 @@ let rec pr_vernac = function
 	| PrintModuleType qid -> str"Print Module Type" ++ spc() ++ pr_reference qid
 	| PrintModule qid -> str"Print Module" ++ spc() ++ pr_reference qid
 	| PrintInspect n -> str"Inspect" ++ spc() ++ int n
-  | PrintSetoids -> str"Print Setoids"
 	| PrintScopes -> str"Print Scopes"
 	| PrintScope s -> str"Print Scope" ++ spc() ++ str s 
 	| PrintVisibility s -> str"Print Visibility" ++ pr_opt str s 
 	| PrintAbout qid -> str"About" ++ spc()  ++ pr_reference qid
 	| PrintImplicit qid -> str"Print Implicit" ++ spc()  ++ pr_reference qid
+(* spiwack: command printing all the axioms and section variables used in a 
+         term *)
+	| PrintAssumptions qid -> str"Print Assumptions"++spc()++pr_reference qid
       in pr_printable p
-  | VernacSearch (sea,sea_r) -> pr_search sea sea_r pr_pattern_expr
+  | VernacSearch (sea,sea_r) -> pr_search sea sea_r pr_constr_pattern_expr
   | VernacLocate loc -> 
       let pr_locate =function
 	| LocateTerm qid ->  pr_reference qid
@@ -843,19 +930,16 @@ and pr_extend s cl =
     let rl = match_vernac_rule (List.map Genarg.genarg_tag cl) rls in
     let start,rl,cl =
       match rl with
-	| Egrammar.TacTerm s :: rl -> str s, rl, cl
-	| Egrammar.TacNonTerm _ :: rl ->
-	    (* Will put an unnecessary extra space in front *)
-	    pr_gen (Global.env()) (List.hd cl), rl, List.tl cl 
-	| [] -> anomaly "Empty entry" in
+      | Egrammar.TacTerm s :: rl -> str s, rl, cl
+      | Egrammar.TacNonTerm _ :: rl -> pr_arg (List.hd cl), rl, List.tl cl 
+      | [] -> anomaly "Empty entry" in
     let (pp,_) =
       List.fold_left
         (fun (strm,args) pi ->
-          match pi with
-              Egrammar.TacNonTerm _ -> 
-                (strm ++ pr_gen (Global.env()) (List.hd args),
-                List.tl args)
-            | Egrammar.TacTerm s -> (strm ++ spc() ++ str s, args))
+          let pp,args = match pi with
+          | Egrammar.TacNonTerm _ -> (pr_arg (List.hd args), List.tl args)
+          | Egrammar.TacTerm s -> (str s, args) in
+          (strm ++ spc() ++ pp), args)
         (start,cl) rl in
     hov 1 pp
   with Not_found ->
