@@ -6,7 +6,7 @@
 (*         *       GNU Lesser General Public License Version 2.1        *)
 (************************************************************************)
 
-(* $Id: logic.ml 9601 2007-02-06 21:37:59Z herbelin $ *)
+(* $Id: logic.ml 11897 2009-02-09 19:28:02Z barras $ *)
 
 open Pp
 open Util
@@ -27,16 +27,17 @@ open Typeops
 open Type_errors
 open Retyping
 open Evarutil
+open Tacexpr
  
 type refiner_error =
 
   (* Errors raised by the refiner *)
   | BadType of constr * constr * constr
-  | OccurMeta of constr
-  | OccurMetaGoal of constr
+  | UnresolvedBindings of name list
   | CannotApply of constr * constr
   | NotWellTyped of constr
   | NonLinearProof of constr
+  | MetaInType of constr
 
   (* Errors raised by the tactics *)
   | IntroNeedsProduct
@@ -48,18 +49,43 @@ open Pretype_errors
 
 let rec catchable_exception = function
   | Stdpp.Exc_located(_,e) -> catchable_exception e
+  | LtacLocated(_,e) -> catchable_exception e
   | Util.UserError _ | TypeError _ 
   | RefinerError _ | Indrec.RecursionSchemeError _
   | Nametab.GlobalizationError _ | PretypeError (_,VarNotFound _)
   (* unification errors *)
-  | PretypeError(_,(CannotUnify _|CannotGeneralize _|NoOccurrenceFound _|
-      CannotUnifyBindingType _|NotClean _)) -> true
+  | PretypeError(_,(CannotUnify _|CannotUnifyLocal _|CannotGeneralize _
+		   |NoOccurrenceFound _|CannotUnifyBindingType _|NotClean _
+		   |CannotFindWellTypedAbstraction _
+		   |UnsolvableImplicit _)) -> true
+  | Typeclasses_errors.TypeClassError 
+      (_, Typeclasses_errors.UnsatisfiableConstraints _) -> true
   | _ -> false
 
+let error_no_such_hypothesis id =
+  error ("No such hypothesis: " ^ string_of_id id ^ ".")
+
 (* Tells if the refiner should check that the submitted rules do not
    produce invalid subgoals *)
 let check = ref false
-let with_check = Options.with_option check
+let with_check = Flags.with_option check
+
+(* [apply_to_hyp sign id f] splits [sign] into [tail::[id,_,_]::head] and
+   returns [tail::(f head (id,_,_) (rev tail))] *)
+let apply_to_hyp sign id f =
+  try apply_to_hyp sign id f 
+  with Hyp_not_found -> 
+    if !check then error "No such assumption."
+    else sign
+
+let apply_to_hyp_and_dependent_on sign id f g =
+  try apply_to_hyp_and_dependent_on sign id f g 
+  with Hyp_not_found -> 
+    if !check then error "No such assumption."
+    else sign
+
+let check_typability env sigma c =
+  if !check then let _ = type_of env sigma c in () 
 
 (************************************************************************)
 (************************************************************************)
@@ -69,30 +95,14 @@ let with_check = Options.with_option check
 (* The Clear tactic: it scans the context for hypotheses to be removed
    (instead of iterating on the list of identifier to be removed, which
    forces the user to give them in order). *)
-let clear_hyps sigma ids gl =
-  let evd = ref (Evd.create_evar_defs sigma) in
-  let ngl = Evarutil.clear_hyps_in_evi evd gl ids in
-    (ngl, evars_of !evd)
+
+let clear_hyps sigma ids sign cl =
+  let evdref = ref (Evd.create_goal_evar_defs sigma) in
+  let (hyps,concl) = Evarutil.clear_hyps_in_evi evdref sign cl ids in
+  (hyps,concl,evars_of !evdref)
 
 (* The ClearBody tactic *)
 
-(* [apply_to_hyp sign id f] splits [sign] into [tail::[id,_,_]::head] and
-   returns [tail::(f head (id,_,_) (rev tail))] *)
-let apply_to_hyp sign id f =
-  try apply_to_hyp sign id f 
-  with Hyp_not_found -> 
-    if !check then error "No such assumption"
-    else sign
-
-let apply_to_hyp_and_dependent_on sign id f g =
-  try apply_to_hyp_and_dependent_on sign id f g 
-  with Hyp_not_found -> 
-    if !check then error "No such assumption"
-    else sign
-
-let check_typability env sigma c =
-  if !check then let _ = type_of env sigma c in () 
-
 let recheck_typability (what,id) env sigma t =
   try check_typability env sigma t
   with _ ->
@@ -107,7 +117,7 @@ let remove_hyp_body env sigma id =
     apply_to_hyp_and_dependent_on (named_context_val env) id
       (fun (_,c,t) _ ->
 	match c with
-	| None -> error ((string_of_id id)^" is not a local definition")
+	| None -> error ((string_of_id id)^" is not a local definition.")
 	| Some c ->(id,None,t))
       (fun (id',c,t as d) sign ->
 	(if !check then
@@ -122,26 +132,117 @@ let remove_hyp_body env sigma id =
   in
   reset_with_named_context sign env 
 
+(* Reordering of the context *)
+
+(* faire le minimum d'echanges pour que l'ordre donne soit un *)
+(* sous-ordre du resultat. Par exemple, 2 hyps non mentionnee ne sont *)
+(* pas echangees. Choix: les hyps mentionnees ne peuvent qu'etre *)
+(* reculees par rapport aux autres (faire le contraire!) *)
+ 
+let mt_q = (Idmap.empty,[])
+let push_val y = function
+    (_,[] as q) -> q
+  | (m, (x,l)::q) -> (m, (x,Idset.add y l)::q)
+let push_item x v (m,l) =
+  (Idmap.add x v m, (x,Idset.empty)::l)
+let mem_q x (m,_) = Idmap.mem x m
+let rec find_q x (m,q) =
+  let v = Idmap.find x m in
+  let m' = Idmap.remove x m in
+  let rec find accs acc = function
+      [] -> raise Not_found
+    | [(x',l)] ->
+        if x=x' then ((v,Idset.union accs l),(m',List.rev acc))
+        else raise Not_found
+    | (x',l as i)::((x'',l'')::q as itl) ->
+        if x=x' then
+          ((v,Idset.union accs l),
+           (m',List.rev acc@(x'',Idset.add x (Idset.union l l''))::q))
+        else find (Idset.union l accs) (i::acc) itl in
+  find Idset.empty [] q
+
+let occur_vars_in_decl env hyps d =
+  if Idset.is_empty hyps then false else
+    let ohyps = global_vars_set_of_decl env d in
+    Idset.exists (fun h -> Idset.mem h ohyps) hyps
+
+let reorder_context env sign ord =
+  let ords = List.fold_right Idset.add ord Idset.empty in
+  if List.length ord <> Idset.cardinal ords then
+    error "Order list has duplicates";
+  let rec step ord expected ctxt_head moved_hyps ctxt_tail =
+    match ord with
+      | [] -> List.rev ctxt_tail @ ctxt_head
+      | top::ord' when mem_q top moved_hyps ->
+          let ((d,h),mh) = find_q top moved_hyps in
+          if occur_vars_in_decl env h d then
+            errorlabstrm "reorder_context"
+              (str "Cannot move declaration " ++ pr_id top ++ spc() ++
+              str "before " ++
+              prlist_with_sep pr_spc pr_id
+                (Idset.elements (Idset.inter h
+                  (global_vars_set_of_decl env d))));
+          step ord' expected ctxt_head mh (d::ctxt_tail)
+      | _ ->
+          (match ctxt_head with
+            | [] -> error_no_such_hypothesis (List.hd ord)
+            | (x,_,_ as d) :: ctxt ->
+                if Idset.mem x expected then
+                  step ord (Idset.remove x expected)
+                    ctxt (push_item x d moved_hyps) ctxt_tail
+                else
+                  step ord expected
+                    ctxt (push_val x moved_hyps) (d::ctxt_tail)) in
+  step ord ords sign mt_q []
+
+let reorder_val_context env sign ord =
+  val_of_named_context (reorder_context env (named_context_of_val sign) ord)
+
+
+
+
+let check_decl_position env sign (x,_,_ as d) =
+  let needed = global_vars_set_of_decl env d in
+  let deps = dependency_closure env (named_context_of_val sign) needed in
+  if List.mem x deps then
+    error ("Cannot create self-referring hypothesis "^string_of_id x);
+  x::deps
+
 (* Auxiliary functions for primitive MOVE tactic
  *
- * [move_after with_dep toleft (left,(hfrom,typfrom),right) hto] moves
- * hyp [hfrom] just after the hyp [hto] which belongs to the hyps on the 
+ * [move_hyp with_dep toleft (left,(hfrom,typfrom),right) hto] moves
+ * hyp [hfrom] at location [hto] which belongs to the hyps on the 
  * left side [left] of the full signature if [toleft=true] or to the hyps 
  * on the right side [right] if [toleft=false].
  * If [with_dep] then dependent hypotheses are moved accordingly. *)
 
+let rec get_hyp_after h = function
+  | [] -> error_no_such_hypothesis h
+  | (hyp,_,_) :: right ->
+      if hyp = h then
+	match right with (id,_,_)::_ -> MoveBefore id | [] -> MoveToEnd false
+      else
+       get_hyp_after h right
+
 let split_sign hfrom hto l =
   let rec splitrec left toleft = function
-    | [] -> error ("No such hypothesis : " ^ (string_of_id hfrom))
+    | [] -> error_no_such_hypothesis hfrom
     | (hyp,c,typ) as d :: right ->
       	if hyp = hfrom then 
-	  (left,right,d,toleft) 
-      	else 
-	  splitrec (d::left) (toleft or (hyp = hto)) right
+	  (left,right,d, toleft or hto = MoveToEnd true)
+      	else
+	  splitrec (d::left) 
+	    (toleft or hto = MoveAfter hyp or hto = MoveBefore hyp)
+	    right
   in 
     splitrec [] false l
 
-let move_after with_dep toleft (left,(idfrom,_,_ as declfrom),right) hto =
+let hyp_of_move_location = function
+  | MoveAfter id -> id 
+  | MoveBefore id -> id
+  | _ -> assert false
+
+let move_hyp with_dep toleft (left,(idfrom,_,_ as declfrom),right) hto =
   let env = Global.env() in
   let test_dep (hyp,c,typ as d) (hyp2,c,typ2 as d2) =
     if toleft
@@ -149,23 +250,27 @@ let move_after with_dep toleft (left,(idfrom,_,_ as declfrom),right) hto =
     else occur_var_in_decl env hyp d2
   in
   let rec moverec first middle = function
-    | [] -> error ("No such hypothesis : " ^ (string_of_id hto))
+    | [] ->
+	if match hto with MoveToEnd _ -> false | _ -> true then
+	  error_no_such_hypothesis (hyp_of_move_location hto);
+	List.rev first @ List.rev middle
+    | (hyp,_,_) :: _ as right when hto = MoveBefore hyp ->
+	List.rev first @ List.rev middle @ right
     | (hyp,_,_) as d :: right ->
 	let (first',middle') =
       	  if List.exists (test_dep d) middle then 
-	    if with_dep & (hyp <> hto) then 
+	    if with_dep & hto <> MoveAfter hyp then 
 	      (first, d::middle)
             else 
-	      error 
-		("Cannot move "^(string_of_id idfrom)^" after "
-		 ^(string_of_id hto)
-		 ^(if toleft then ": it occurs in " else ": it depends on ")
-		 ^(string_of_id hyp))
-          else 
+	      errorlabstrm "" (str "Cannot move " ++ pr_id idfrom ++
+	        pr_move_location pr_id hto ++ 
+	        str (if toleft then ": it occurs in " else ": it depends on ")
+	        ++ pr_id hyp ++ str ".")
+          else
 	    (d::first, middle)
 	in
-      	if hyp = hto then 
-	  (List.rev first')@(List.rev middle')@right
+      	if hto = MoveAfter hyp then
+	  List.rev first' @ List.rev middle' @ right
       	else 
 	  moverec first' middle' right
   in
@@ -181,53 +286,11 @@ let move_after with_dep toleft (left,(idfrom,_,_ as declfrom),right) hto =
     List.fold_left (fun sign d -> push_named_context_val d sign)
       right left
 
-let check_backward_dependencies sign d =
-  if not (Idset.for_all
-	    (fun id -> mem_named_context id sign)
-	    (global_vars_set_of_decl (Global.env()) d))
-  then
-    error "Can't introduce at that location: free variable conflict"
-
-
-let check_forward_dependencies id tail =
-  let env = Global.env() in
-  List.iter
-    (function (id',_,_ as decl) ->
-       if occur_var_in_decl env id decl then
-	 error ((string_of_id id) ^ " is used in hypothesis " 
-		^ (string_of_id id')))
-    tail
-
-let check_goal_dependency id cl =
-  let env = Global.env() in
-    if Idset.mem id (global_vars_set env cl) then
-      error (string_of_id id^" is used in conclusion")
-
 let rename_hyp id1 id2 sign =
   apply_to_hyp_and_dependent_on sign id1
     (fun (_,b,t) _ -> (id2,b,t))
     (fun d _ -> map_named_declaration (replace_vars [id1,mkVar id2]) d)
 
-let replace_hyp sign id d cl =
-  if !check then
-    check_goal_dependency id cl;
-  apply_to_hyp sign id
-    (fun sign _ tail ->
-      if !check then
-	(check_backward_dependencies sign d;
-	 check_forward_dependencies id tail); 
-      d)
-
-(* why we dont check that id does not appear in tail ??? *)
-let insert_after_hyp sign id d =
-  try 
-    insert_after_hyp sign id d 
-      (fun sign ->
-	if !check then check_backward_dependencies sign d)
-  with Hyp_not_found ->
-    if !check then error "No such assumption"
-    else sign
-
 (************************************************************************)
 (************************************************************************)
 (* Implementation of the logical rules *)
@@ -235,14 +298,23 @@ let insert_after_hyp sign id d =
 (* Will only be used on terms given to the Refine rule which have meta 
 variables only in Application and Case *)
 
+let error_unsupported_deep_meta c =
+  errorlabstrm "" (strbrk "Application of lemmas whose beta-iota normal " ++
+    strbrk "form contains metavariables deep inside the term is not " ++
+    strbrk "supported; try \"refine\" instead.")
+
 let collect_meta_variables c = 
-  let rec collrec acc c = match kind_of_term c with
-    | Meta mv -> mv::acc
-    | Cast(c,_,_) -> collrec acc c
-    | (App _| Case _) -> fold_constr collrec acc c
-    | _ -> acc
-  in 
-  List.rev(collrec [] c)
+  let rec collrec deep acc c = match kind_of_term c with
+    | Meta mv -> if deep then error_unsupported_deep_meta () else mv::acc
+    | Cast(c,_,_) -> collrec deep acc c
+    | (App _| Case _) -> fold_constr (collrec deep) acc c
+    | _ -> fold_constr (collrec true) acc c
+  in
+  List.rev (collrec false [] c)
+
+let check_meta_variables c = 
+  if not (list_distinct (collect_meta_variables c)) then
+    raise (RefinerError (NonLinearProof c))
 
 let check_conv_leq_goal env sigma arg ty conclty =
   if !check & not (is_conv_leq env sigma ty conclty) then 
@@ -264,9 +336,10 @@ let rec mk_refgoals sigma goal goalacc conclty trm =
 *)
   match kind_of_term trm with
     | Meta _ ->
-	if occur_meta conclty then
-          raise (RefinerError (OccurMetaGoal conclty));
-	(mk_goal hyps (nf_betaiota conclty))::goalacc, conclty
+	let conclty = nf_betaiota sigma conclty in
+	  if !check && occur_meta conclty then
+	    raise (RefinerError (MetaInType conclty));
+	  (mk_goal hyps conclty)::goalacc, conclty
 
     | Cast (t,_, ty) ->
 	check_typability env sigma ty;
@@ -276,11 +349,11 @@ let rec mk_refgoals sigma goal goalacc conclty trm =
     | App (f,l) ->
 	let (acc',hdty) =
 	  match kind_of_term f with
-	    | (Ind _ (* needed if defs in Type are polymorphic: | Const _*))
-		when not (array_exists occur_meta l) (* we could be finer *) ->
+	    | Ind _ | Const _
+		when (isInd f or has_polymorphic_type (destConst f)) ->
 		(* Sort-polymorphism of definition and inductive types *)
 		goalacc, 
-		type_of_global_reference_knowing_parameters env sigma f l
+                type_of_global_reference_knowing_conclusion env sigma f conclty
 	    | _ -> 
 		mk_hdgoals sigma goal goalacc f
 	in
@@ -300,7 +373,9 @@ let rec mk_refgoals sigma goal goalacc conclty trm =
 	(acc'',conclty')
 
     | _ -> 
-	if occur_meta trm then raise (RefinerError (OccurMeta trm));
+	if occur_meta trm then
+	  anomaly "refiner called with a meta in non app/case subterm";
+
       	let t'ty = goal_type_of env sigma trm in
 	check_conv_leq_goal env sigma trm t'ty conclty;
         (goalacc,t'ty)
@@ -315,7 +390,7 @@ and mk_hdgoals sigma goal goalacc trm =
   match kind_of_term trm with
     | Cast (c,_, ty) when isMeta c ->
 	check_typability env sigma ty;
-	(mk_goal hyps (nf_betaiota ty))::goalacc,ty
+	(mk_goal hyps (nf_betaiota sigma ty))::goalacc,ty
 
     | Cast (t,_, ty) ->
 	check_typability env sigma ty;
@@ -340,7 +415,10 @@ and mk_hdgoals sigma goal goalacc trm =
 	in
 	(acc'',conclty')
 
-    | _ -> goalacc, goal_type_of env sigma trm
+    | _ ->
+	if !check && occur_meta trm then
+	  anomaly "refined called with a dependent meta";
+	goalacc, goal_type_of env sigma trm
 
 and mk_arggoals sigma goal goalacc funty = function
   | [] -> goalacc,funty
@@ -367,20 +445,20 @@ and mk_casegoals sigma goal goalacc p c =
   (acc'',lbrty,conclty)
 
 
-let error_use_instantiate () =
-  errorlabstrm "Logic.prim_refiner"
-    (str"cannot intro when there are open metavars in the domain type" ++
-       spc () ++ str"- use Instantiate")
-
 let convert_hyp sign sigma (id,b,bt as d) =
-  apply_to_hyp sign id
-    (fun _ (_,c,ct) _ ->
-       let env = Global.env_of_context sign in
-       if !check && not (is_conv env sigma bt ct) then
-	 error ("Incorrect change of the type of "^(string_of_id id));
-       if !check && not (option_compare (is_conv env sigma) b c) then
-	 error ("Incorrect change of the body of "^(string_of_id id));
-       d)
+  let env = Global.env() in
+  let reorder = ref [] in
+  let sign' =
+    apply_to_hyp sign id
+      (fun _ (_,c,ct) _ ->
+        let env = Global.env_of_context sign in
+        if !check && not (is_conv env sigma bt ct) then
+	  error ("Incorrect change of the type of "^(string_of_id id)^".");
+        if !check && not (Option.Misc.compare (is_conv env sigma) b c) then
+	  error ("Incorrect change of the body of "^(string_of_id id)^".");
+       if !check then reorder := check_decl_position env sign d;
+       d) in
+  reorder_val_context env sign' !reorder
 
 (* Normalizing evars in a goal. Called by tactic Local_constraints
    (i.e. when the sigma of the proof tree changes). Detect if the
@@ -412,12 +490,10 @@ let prim_refiner r sigma goal =
 	  error "New variable is already declared";
         (match kind_of_term (strip_outer_cast cl) with
 	   | Prod (_,c1,b) ->
-	       if occur_meta c1 then error_use_instantiate();
 	       let sg = mk_goal (push_named_context_val (id,None,c1) sign)
 			  (subst1 (mkVar id) b) in
 	       ([sg], sigma)
 	   | LetIn (_,c1,t1,b) ->
-	       if occur_meta c1 or occur_meta t1 then error_use_instantiate();
 	       let sg =
 		 mk_goal (push_named_context_val (id,Some c1,t1) sign)
 		   (subst1 (mkVar id) b) in
@@ -425,28 +501,21 @@ let prim_refiner r sigma goal =
 	   | _ ->
 	       raise (RefinerError IntroNeedsProduct))
 	
-    | Intro_replacing id ->
-	(match kind_of_term (strip_outer_cast cl) with
-           | Prod (_,c1,b) ->
-	       if occur_meta c1 then error_use_instantiate();
-	       let sign' = replace_hyp sign id (id,None,c1) cl in
-	       let sg = mk_goal sign' (subst1 (mkVar id) b) in
-	       ([sg], sigma)
-           | LetIn (_,c1,t1,b) ->
-	       if occur_meta c1 then error_use_instantiate();
-	       let sign' = replace_hyp sign id (id,Some c1,t1) cl in
-	       let sg = mk_goal sign' (subst1 (mkVar id) b) in
-	       ([sg], sigma)
-	   | _ ->
-	       raise (RefinerError IntroNeedsProduct))
-	
-    | Cut (b,id,t) ->
-    	if !check && mem_named_context id (named_context_of_val sign) then
-	  error "New variable is already declared";
-        if occur_meta t then error_use_instantiate();
-        let sg1 = mk_goal sign (nf_betaiota t) in
-        let sg2 = mk_goal (push_named_context_val (id,None,t) sign) cl in
-        if b then ([sg1;sg2],sigma) else ([sg2;sg1], sigma)
+    | Cut (b,replace,id,t) ->
+        let sg1 = mk_goal sign (nf_betaiota sigma t) in
+	let sign,cl,sigma =
+	  if replace then
+	    let nexthyp = get_hyp_after id (named_context_of_val sign) in
+	    let sign,cl,sigma = clear_hyps sigma [id] sign cl in
+	    move_hyp true false ([],(id,None,t),named_context_of_val sign)
+	      nexthyp,
+	      cl,sigma
+	  else
+            (if !check && mem_named_context id (named_context_of_val sign) then
+                error "New variable is already declared";
+	     push_named_context_val (id,None,t) sign,cl,sigma) in
+        let sg2 = mk_goal sign cl in
+        if b then ([sg1;sg2],sigma) else ([sg2;sg1],sigma)
 
     | FixRule (f,n,rest) ->
      	let rec check_ind env k cl = 
@@ -456,10 +525,10 @@ let prim_refiner r sigma goal =
 		  try 
 		    fst (find_inductive env sigma c1)
 		  with Not_found -> 
-		    error "cannot do a fixpoint on a non inductive type"
+		    error "Cannot do a fixpoint on a non inductive type."
             	else 
 		  check_ind (push_rel (na,None,c1) env) (k-1) b
-            | _ -> error "not enough products"
+            | _ -> error "Not enough products."
 	in
 	let (sp,_) = check_ind env n cl in
 	let all = (f,n,cl)::rest in
@@ -467,10 +536,10 @@ let prim_refiner r sigma goal =
 	  | (f,n,ar)::oth ->
 	      let (sp',_)  = check_ind env n ar in 
 	      if not (sp=sp') then 
-		error ("fixpoints should be on the same " ^ 
-		       "mutual inductive declaration");
+		error ("Fixpoints should be on the same " ^ 
+		       "mutual inductive declaration.");
 	      if !check && mem_named_context f (named_context_of_val sign) then
-		error "name already used in the environment";
+		error "Name already used in the environment";
 	      mk_sign (push_named_context_val (f,None,ar) sign) oth
 	  | [] -> 
 	      List.map (fun (_,_,c) -> mk_goal sign c) all
@@ -487,8 +556,7 @@ let prim_refiner r sigma goal =
 		  let _ = find_coinductive env sigma b in ()
                 with Not_found -> 
 		  error ("All methods must construct elements " ^
-			  "in coinductiv-> goal
-e types")
+			  "in coinductive types.")
 	in
 	let all = (f,cl)::others in
      	List.iter (fun (_,c) -> check_is_coind env c) all;
@@ -496,7 +564,7 @@ e types")
           | (f,ar)::oth ->
 	      (try
                 (let _ = lookup_named_val f sign in
-                error "name already used in the environment")
+                error "Name already used in the environment.")
               with
               |	Not_found ->
                   mk_sign (push_named_context_val (f,None,ar) sign) oth)
@@ -505,8 +573,7 @@ e types")
 	  (mk_sign sign all, sigma)
 
     | Refine c ->
-        if not (list_distinct (collect_meta_variables c)) then
-          raise (RefinerError (NonLinearProof c));
+	check_meta_variables c;
 	let (sgl,cl') = mk_refgoals sigma goal [] cl c in
 	let sgl = List.rev sgl in
 	  (sgl, sigma)
@@ -525,8 +592,8 @@ e types")
 
     (* And now the structural rules *)
     | Thin ids -> 
-	let (ngl, nsigma) = clear_hyps sigma ids goal in
-	  ([ngl], nsigma)
+	let (hyps,concl,nsigma) = clear_hyps sigma ids sign cl in
+	  ([mk_goal hyps concl], nsigma)
 	
     | ThinBody ids ->
 	let clear_aux env id =
@@ -542,13 +609,17 @@ e types")
   	let (left,right,declfrom,toleft) = 
 	  split_sign hfrom hto (named_context_of_val sign) in
   	let hyps' = 
-	  move_after withdep toleft (left,declfrom,right) hto in
+	  move_hyp withdep toleft (left,declfrom,right) hto in
   	  ([mk_goal hyps' cl], sigma)
 
+    | Order ord ->
+        let hyps' = reorder_val_context env sign ord in
+        ([mk_goal hyps' cl], sigma)
+
     | Rename (id1,id2) ->
         if !check & id1 <> id2 &&
 	  List.mem id2 (ids_of_named_context (named_context_of_val sign)) then
-          error ((string_of_id id2)^" is already used");
+          error ((string_of_id id2)^" is already used.");
         let sign' = rename_hyp id1 id2 sign in
         let cl' = replace_vars [id1,mkVar id2] cl in
           ([mk_goal sign' cl'], sigma)
@@ -582,10 +653,10 @@ let subst_proof_vars =
          
 let rec rebind id1 id2 = function
   | [] -> [Name id2,SectionVar id1]
-  | (na,_ as x)::l -> 
-      if na = Name id1 then (Name id2,ProofVar)::l else
+  | (na,k as x)::l -> 
+      if na = Name id1 then (Name id2,k)::l else
         let l' = rebind id1 id2 l in
-        if na = Name id2 then (Anonymous,ProofVar)::l' else x::l'
+        if na = Name id2 then (Anonymous,k)::l' else x::l'
 
 let add_proof_var id vl = (Name id,ProofVar)::vl
 
@@ -609,20 +680,9 @@ let prim_extractor subfun vl pft =
 		let cb = subst_proof_vars vl b in
 		let cty = subst_proof_vars vl ty in
 		  mkLetIn (Name id, cb, cty, subfun (add_proof_var id vl) spf)
-	    | _ -> error "incomplete proof!")
+	    | _ -> error "Incomplete proof!")
 	    
-      | Some (Prim (Intro_replacing id),[spf]) ->
-	  (match kind_of_term (strip_outer_cast cl) with
-	    | Prod (_,ty,_) ->
-		let cty = subst_proof_vars vl ty in
-		  mkLambda (Name id, cty, subfun (add_proof_var id vl) spf)
-	    | LetIn (_,b,ty,_) ->
-		let cb = subst_proof_vars vl b in
-		let cty = subst_proof_vars vl ty in
-		  mkLetIn (Name id, cb, cty, subfun (add_proof_var id vl) spf)
-	    | _ -> error "incomplete proof!")
-
-      | Some (Prim (Cut (b,id,t)),[spf1;spf2]) ->
+      | Some (Prim (Cut (b,_,id,t)),[spf1;spf2]) ->
           let spf1, spf2 = if b then spf1, spf2 else spf2, spf1 in
 	    mkLetIn (Name id,subfun vl spf1,subst_proof_vars vl t,
                     subfun (add_proof_var id vl) spf2)
@@ -666,7 +726,7 @@ let prim_extractor subfun vl pft =
       | Some (Prim (ThinBody _),[pf]) ->
 	  subfun vl pf
 	    
-      | Some (Prim (Move _),[pf]) ->
+      | Some (Prim (Move _|Order _),[pf]) ->
 	  subfun vl pf
 
       | Some (Prim (Rename (id1,id2)),[pf]) ->
